from typing import Union, Iterable

import nip.elements
from .parser import parse


def update_flatten(
    base_config: "nip.elements.Node", updating_config: Union[str, "nip.elements.Node"]
) -> "nip.elements.Node":
    """Updates base_config with flatten params described in other_config. Inplace.

    Notes
    -----
    updating_config is expected to be flat dict and deep nodes of base_config are accessed via dots in keys.
    (e.g. `some.deep.node: new_value`)

    Parameters
    ----------
    base_config : str or Node
        Basic config Node to be updated.
        If str, will be treated as a path to config.
    updating_config : Node
        Config that will be used to update the basic one.

    Returns
    -------
    updated_config : Node
        Updated config Node.
    """
    if isinstance(updating_config, str):
        updating_config = parse(updating_config)
    if isinstance(updating_config, Iterable):
        raise TypeError(
            "Only non iterable configs are supported as updating_config."
        )  # mb: add IterableConfig, instead of handling this by Parser.
    if isinstance(updating_config, nip.elements.Document):
        updating_config = updating_config._value
    if not isinstance(updating_config, nip.elements.Args) or not updating_config._is_dict():
        raise TypeError(f"Flatten updating config should be just a dict.")
    for key, value in updating_config:
        base_config[key] = value
    base_config._get_root()._update_parents()
    return base_config


def _update(base_config: "nip.elements.Node", updating_config: "nip.elements.Node"):
    if isinstance(
        base_config, (nip.elements.Document, nip.elements.Link, nip.elements.LinkCreation, nip.elements.Tag)
    ):  # step into
        base_config._value = _update(base_config._value, updating_config)
        return base_config
    if not isinstance(base_config, nip.elements.Args) or not isinstance(
        updating_config, nip.elements.Args
    ):  # leaf in any conf
        return updating_config
    for key, value in updating_config:  # both args and kwargs
        if key in base_config:
            base_config[key] = _update(base_config[key], value)
        else:
            base_config[key] = value  # adding args via next index
    return base_config


def update(base_config: "nip.elements.Node", updating_config: "nip.elements.Node") -> "nip.elements.Node":  # mb: merge
    """Updates base_config with other_config recursively traversing both trees. Inplace.

    Notes
    -----
    Any leaf Node of base_config will be overwritten by corresponding Node of updating_config.
    Any Node of updating_config that is not dict or list will overwrite corresponding Node of base_config.
    It is not possible to overwrite dict nodes atm, since they will be simply updated.

    Parameters
    ----------
    base_config : Node
        Basic config Node to be updated.
    updating_config : Node
        Config that will be used to update the basic one.

    Returns
    -------
    updated_config : Node
        Updated config Node.
    """
    if isinstance(updating_config, str):
        updating_config = parse(updating_config)
    if isinstance(updating_config, nip.elements.Document):
        updating_config = updating_config._value
    base_config = _update(base_config, updating_config)
    base_config._get_root()._update_parents()
    return base_config



"""Contains nip directives."""

import nip.elements
from nip.parser import Parser, ParserError, parse
from .stream import Stream
from .update import update


def insert_directive(node, stream: Stream):  # mb: class; mb: just a Node, not a directive
    if isinstance(node, nip.elements.Value):
        path = node.to_python()
        assert isinstance(path, str), "Load directive expects path as an argument."
        parser = Parser()
        config = parser.parse(path)  # Document
        return config._value

    elif isinstance(node, nip.elements.Args):
        assert len(node._args) == 1, "only single positional argument will be treated as config path."
        path = node[0].to_python()
        assert isinstance(path, str), "Load directive expects path as first argument."
        parser = Parser()
        parser.link_replacements = node._kwargs
        config = parser.parse(path)  # Document
        return config._value

    else:
        raise ParserError(
            stream,
            "string or combination of arg and **kwargs are expected as value of !!insert directive",
        )


def update_directive(node, stream: Stream):  # mb: skip directive operator
    if not isinstance(node, nip.elements.Args) or "_update_" not in node:
        raise ParserError(stream, "!!update directive expects dict node with __update__ key.")
    update_config_path = node["_update_"].to_python()
    del node._kwargs["_update_"]
    return update(node, update_config_path)


def base_directive(node, stream: Stream):  # mb: parent
    if not isinstance(node, nip.elements.Args) or "_base_" not in node:
        raise ParserError(stream, "!!base directive expects dict node with __base__ key.")
    base_config_path = node["_base_"].to_python()
    base_config = parse(base_config_path)
    if isinstance(base_config, nip.elements.Document):
        base_config = base_config._value
    del node._kwargs["_base_"]
    return update(base_config, node)


_directives = {"insert": insert_directive, "update": update_directive, "base": base_directive}


def call_directive(name, right_value, stream: Stream):
    if name not in _directives:
        raise ParserError(stream, f"Unknown parser directive '{name}'.")
    return _directives[name](right_value, stream)



from typing import Optional, Callable, Union

from nip.constructor import global_builders, ArgsKwargs
from nip.elements import Node, Args, Tag, Value

global_convertors = {}


class Convertor:
    def __init__(self, load_globals: bool = True):
        self.convertors = {}
        if load_globals:
            self._load_globals()

    def convert(self, obj: object) -> Node:
        class_name = type(obj).__name__
        if class_name in self.convertors:
            tag, convertor = self.convertors[class_name]
            return Tag(tag, self.convert(convertor(obj)))

        if hasattr(obj, "__nip__"):
            return Tag(class_name, self.convert(obj.__nip__()))

        if isinstance(obj, ArgsKwargs):
            args = [self.convert(value) for value in obj.args]
            kwargs = {key: self.convert(value) for key, value in obj.kwargs.items()}
            return Args(args=args, kwargs=kwargs)

        if isinstance(obj, dict):
            kwargs = {}
            for key, value in obj.items():
                kwargs[key] = self.convert(value)
            return Args(args=[], kwargs=kwargs)

        if isinstance(obj, (list, tuple)):
            return Args(args=[self.convert(value) for value in obj], kwargs={})

        if isinstance(obj, (int, float, str, bool, type(None))):
            return Value("value", obj)

        raise ConvertorError(obj, "No convertor specified for this class")

    def register(self, class_: Union[type, str], func: Callable, tag: Optional[str] = None):
        if isinstance(class_, type):
            class_name = class_.__name__
        elif isinstance(class_, str):
            class_name = class_
        else:
            raise TypeError("Expected type or str as class_ argument")
        if tag is None:
            for builder_tag, builder in global_builders.items():
                if isinstance(builder, type) and builder.__class__.__name__ == class_name:
                    tag = builder_tag
            tag = tag or class_name
        self.convertors[class_name] = (tag, func)

    def _load_globals(self):
        self.convertors.update(global_convertors)
        for builder_tag, builder in global_builders.items():
            if isinstance(builder, type) and hasattr(builder, "__nip__"):
                self.convertors[builder.__name__] = (builder_tag, builder.__nip__)


class ConvertorError(Exception):
    def __init__(self, obj, message: str):
        self.class_name = obj.__class__.__name__
        self.obj = obj
        self.message = message

    def __str__(self):
        return f"Unable to convert object {self.obj} of class {self.class_name} to nip: " f"{self.message}"


def pin(class_name: str, tag: str):
    """Wrapper that registers the function as an object dumper.

    Parameters
    ----------
    class_name:
        Class that should be dumped using this function.
    tag:
        What tag should be used for this class in config file.
    """
    assert tag is None or len(tag) > 0, "name should be nonempty"

    def _(item: type):
        if not isinstance(item, type):
            raise ValueError("Expected class type to be wrapped.")
        global_convertors[class_name] = (tag, item)
        return item

    return _



from pathlib import Path
from typing import Union, Any, Iterable, Callable, Optional

from nip.parser.parse_func import parse, parse_string
from nip.update import update_flatten, update
from . import elements
from .constructor import Constructor
from .convertor import Convertor
from .dumper import Dumper
from .non_seq_constructor import NonSequentialConstructor
from .dict import DictObject

__all__ = [
    "parse",
    "parse_string",
    "construct",
    "load",
    "load_string",
    "run",
    "dump",
    "dump_string",
    "convert",
    "update",
    "update_flatten",
]


def construct(
    config: elements.Node,
    base_config: elements.Node = None,
    strict_typing: bool = False,
    nonsequential: bool = True,
    as_dictobj: bool = True,
) -> Any:
    """Constructs python object based on config and known nip-objects

    Parameters
    ----------
    config: Element
        Config node to be constructed.
    base_config:
        Base config, in case of construction a part of a config with external links.
    strict_typing:
        If True, raises Exception when typing mismatch.
    nonsequential:
        If True, allows to use links before creation.
        Always true if base_config is specified.
    as_dictobj:
        Returns DictObj if true and possible

    Returns
    -------
    obj: Any
    """
    if nonsequential or base_config is not None:
        base_config = base_config or config._get_root()
        constructor = NonSequentialConstructor(base_config, strict_typing=strict_typing, as_dictobj=as_dictobj)
    else:
        constructor = Constructor(strict_typing=strict_typing, as_dictobj=as_dictobj)
    result = constructor.construct(config)
    return result


def _iter_load(configs, strict_typing, nonsequential, as_dictobj):  # Otherwise load() will always be an iterator
    for config in configs:
        yield construct(config, strict_typing=strict_typing, nonsequential=nonsequential, as_dictobj=as_dictobj)


def load(
    path: Union[str, Path],
    always_iter: bool = False,
    strict: bool = False,
    nonsequential: bool = True,
    as_dictobj: bool = False,
) -> Union[Any, Iterable[Any]]:
    """Parses config and constructs python object
    Parameters
    ----------
    path: str or Path
        Path to config file
    always_iter: bool
        If True will always return iterator over configs.
    strict:
        If True, raises Exception when typing mismatch or overwriting dict key.
    nonsequential:
        If True, allows to use links before creation.
    as_dictobj:
        Returns DictObj if true and possible

    Returns
    -------
    obj: Any or Iterable[Any]
    """
    config = parse(path, always_iter, strict=strict)

    if isinstance(config, Iterable):
        return _iter_load(config, strict, nonsequential, as_dictobj)

    return construct(config, strict_typing=strict, nonsequential=nonsequential, as_dictobj=as_dictobj)


def load_string(
    config_string: str,
    always_iter: bool = False,
    strict: bool = False,
    nonsequential: bool = True,
    as_dictobj: bool = True,
) -> Union[Any, Iterable[Any]]:
    """Parses config and constructs python object
    Parameters
    ----------
    config_string: str
        Config as a string.
    always_iter: bool
        If True will always return iterator over configs.
    strict:
        If True, raises Exception when typing mismatch or overwriting dict key.
    nonsequential:
        If True, allows to use links before creation.
    as_dictobj:
        Returns DictObj if true and possible

    Returns
    -------
    obj: Any or Iterable[Any]
    """
    config = parse_string(config_string, always_iter, strict=strict)

    if isinstance(config, Iterable):
        return _iter_load(config, strict, nonsequential, as_dictobj)

    return construct(config, strict_typing=strict, nonsequential=nonsequential, as_dictobj=as_dictobj)


def dump(path: Union[str, Path], obj: Union[elements.Node, object]):
    """Dumps config tree to file.

    Parameters
    ----------
    path: str or Path
        Path to save the config.
    obj: Element or object
        Read or generated config if Element. In case of any other object `convert` will be called.
    """
    if not isinstance(obj, elements.Node):
        obj = convert(obj)  # mb: wrap with Document to ensure getting `---` at the beginning of the file.
    dumper = Dumper()
    dumper.dump(path, obj)


def dump_string(obj: Union[elements.Node, object]) -> str:
    """Dumps config tree to string.

    Parameters
    ----------
    obj: Element
        Read or generated config tree.

    Returns
    -------
    string: str
        Dumped element as a string.
    """
    if not isinstance(obj, elements.Node):
        obj = convert(obj)
    dumper = Dumper()
    return dumper.dumps(obj)


def convert(obj):
    """Converts object to nip.Element

    Parameters
    ----------
    obj: nip-serializable object.

    Returns
    -------
    Element:
        Object representation as Nip.Element

    Notes
    -----
    Any standard python objects that can be casted to and from nip are supported.
    For custom classes method `__nip__` that casts them to the standard object are needed.
    Anything that is returned from `__nip__` method will be recursively converted.
    `__nip__` method is expected to return everything needed for object construction with __init__
    or any other used constructor.
    By default, tag specified in @nip or default class name otherwise will be used as a tag.
    """
    convertor = Convertor()
    return convertor.convert(obj)


def _run_return(value, config, return_values, return_configs):
    return_tuple = []
    if return_values:
        return_tuple.append(value)
    if return_configs:
        return_tuple.append(config)
    if len(return_tuple) == 0:
        return None
    if len(return_tuple) == 1:
        return return_tuple[0]
    return tuple(return_tuple)


def _single_run(
    config, func, verbose, return_values, return_configs, config_parameter, strict, nonsequential, as_dictobj
):
    if verbose:
        print("=" * 20)
        print("Running config:")
        print(dump_string(config))
        print("----")

    value = construct(config, strict_typing=strict, nonsequential=nonsequential, as_dictobj=as_dictobj)
    if func is not None:
        if isinstance(value, tuple) and isinstance(value[0], list) and isinstance(value[1], dict):
            args, kwargs = value
        elif isinstance(value, list):
            args, kwargs = value, {}
        elif isinstance(value, dict):
            args, kwargs = [], value
        else:
            raise RuntimeError("Value constructed by the config cant be parsed as args and kwargs")

        if config_parameter:
            if config_parameter in kwargs:
                raise RuntimeWarning(
                    f"nip.run() was asked to add config parameter '{config_parameter}', "
                    f"but it is already specified by the config. It will be overwritten."
                )
            kwargs[config_parameter] = config

        value = func(*args, **kwargs)

    if verbose:
        print("----")
        print("Run value:")
        print(value)

    return _run_return(value, config, return_values, return_configs)


def _iter_run(
    configs, func, verbose, return_values, return_configs, config_parameter, strict, nonsequential, as_dictobj
):
    for config in configs:
        run_return = _single_run(
            config, func, verbose, return_values, return_configs, config_parameter, strict, nonsequential, as_dictobj
        )
        if run_return:
            yield run_return


def run(
    path: Union[str, Path],
    func: Optional[Callable] = None,
    verbose: bool = True,
    strict: bool = False,
    nonsequential: bool = False,
    return_values: bool = True,
    return_configs: bool = False,
    always_iter: bool = False,
    config_parameter: Optional[str] = None,
    as_dictobj: bool = False,
):
    """Runs config. Config should be declared with function to run as a tag for the Document.
    In case of iterable configs we will iterate over and run each of them.

    Parameters
    ----------
    path: str or Path
        path to config to run.
    func:
        Function to be called with loaded configs.
        If not specified, config will be constructed as is.
    verbose: bool, optional
        Whether to print information about currently running experiment or not.
    strict:
        If True, raises Exception when typing mismatch or overwriting dict key.
    nonsequential:
        If True, allows to use links before creation.
    return_values: bool, optional
        Whether to return values of ran functions or not.
    return_configs: bool, optional
        Whether to return config for ran function or not.
    always_iter: bool, optional
        Result will always be iterable.
    config_parameter: str, optional
        If specified, parsed config will be passed to called function as a parameter with this name.
        `func` parameter must be specified.
    as_dictobj:
        Returns DictObj if true and possible

    Returns
    -------
    Iterator or single value depending on whether the config is iterable
    value or config or (value, config): Any or str or (Any, str)
        returned values and configs of runs
    """
    assert (
        config_parameter is None or config_parameter is not None and func is not None
    ), "`config_parameter` can be used only with specified `func`"
    config = parse(path, always_iter=always_iter, strict=strict)
    if isinstance(config, Iterable):
        return list(
            _iter_run(
                configs=config,
                func=func,
                verbose=verbose,
                return_values=return_values,
                return_configs=return_configs,
                config_parameter=config_parameter,
                strict=strict,
                nonsequential=nonsequential,
                as_dictobj=as_dictobj,
            )
        )  # mb iter?

    return _single_run(
        config=config,
        func=func,
        verbose=verbose,
        return_values=return_values,
        return_configs=return_configs,
        config_parameter=config_parameter,
        strict=strict,
        nonsequential=nonsequential,
        as_dictobj=as_dictobj,
    )



from pathlib import Path
from typing import Union, Iterable

from nip import elements
from nip.parser.iter_parser import IterParser
from nip.parser.parser import Parser


def parse(
    path: Union[str, Path],
    always_iter: bool = False,
    implicit_fstrings: bool = True,
    strict: bool = False,
) -> Union["elements.Node", "Iterable[elements.Node]"]:
    """Parses config providing Element tree

    Parameters
    ----------
    path: str or Path
        path to config file
    always_iter: bool
        If True will always return iterator over configs.
    implicit_fstrings: boot, default: True
        If True, all quoted strings will be treated as python f-strings.
    strict:
        It True, checks overwriting dict keys and positioning (`args` before `kwargs`).

    Returns
    -------
    tree: Element or Iterable[Element]
    """
    parser = Parser(implicit_fstrings=implicit_fstrings, strict=strict)
    tree = parser.parse(path)
    if parser.has_iterators() or always_iter:
        return IterParser(parser).iter_configs(tree)
    return tree


def parse_string(
    config_string: str,
    always_iter: bool = False,
    implicit_fstrings: bool = True,
    strict: bool = False,
) -> Union["elements.Node", "Iterable[elements.Node]"]:
    """Parses config providing Element tree

    Parameters
    ----------
    config_string: str
        Config as a string.
    always_iter: bool
        If True will always return iterator over configs.
    implicit_fstrings: boot, default: True
        If True, all quoted strings will be treated as python f-strings.
    strict:
        It True, checks overwriting dict keys and positioning (`args` before `kwargs`).

    Returns
    -------
    tree: Element or Iterable[Element]
    """
    parser = Parser(implicit_fstrings=implicit_fstrings, strict=strict)
    tree = parser.parse_string(config_string)
    if parser.has_iterators() or always_iter:
        return IterParser(parser).iter_configs(tree)
    return tree



import ast
import inspect
from typing import List, Dict, Union, Any

import typeguard


def get_subclasses(cls):
    return set(cls.__subclasses__()).union([s for c in cls.__subclasses__() for s in get_subclasses(c)])


def get_sub_dict(base_cls):
    return {cls.__name__: cls for cls in get_subclasses(base_cls)}


def deep_equals(first: Union[Dict, List, Any], second: Union[Dict, List, Any]):
    if type(first) is type(second):
        return False
    if isinstance(first, dict):
        for key in first:
            if key not in second:
                return False
            if not deep_equals(first[key], second[key]):
                return False
    elif isinstance(first, list):
        if len(first) != len(second):
            return False
        for i, j in zip(first, second):
            if not deep_equals(i, j):
                return False
    else:
        return first == second
    return True


def iterate_items(obj):
    if isinstance(obj, (list, tuple)):
        for i, value in enumerate(obj):
            yield i, value
    if isinstance(obj, dict):
        for key, value in obj.items():
            yield key, value


def flatten(obj, delimiter=".", keys=()):
    if not isinstance(obj, (list, tuple, dict)):
        return {delimiter.join(keys): obj}

    result = {}
    for key, value in iterate_items(obj):
        result.update(flatten(value, delimiter, keys=keys + (str(key),)))
    return result


def check_typing(func, args, kwargs) -> List[str]:
    try:
        signature = inspect.signature(func)
    except ValueError:  # unable to load function signature (common case for builtin functions)
        return []
    messages = []
    typeguard.typechecked()
    if len(args) > len(signature.parameters.values()):
        messages.append("Too many arguments")
    for arg, param in zip(args, signature.parameters.values()):
        print(arg, param)
        if param.annotation is inspect.Parameter.empty:
            continue
        try:
            typeguard.check_type(arg, param.annotation)
        except typeguard.TypeCheckError as e:
            messages.append(f"{param.name}: {e}")

    for name, value in kwargs.items():
        if name not in signature.parameters:
            continue  # handled by python
        annotation = signature.parameters[name].annotation
        # print(name, annotation, value)
        if annotation is inspect.Parameter.empty:
            continue
        try:
            typeguard.check_type(value, annotation)
        except typeguard.TypeCheckError as e:
            messages.append(f"{name}: {e}")

    return messages


class Namespace:
    def __getattr__(self, key: str) -> "Namespace":
        item = Namespace()
        setattr(self, key, item)
        return item

    def __setitem__(self, key, value):
        prefix = key.split(".")[0]
        suffix = ".".join(key.split(".")[1:])
        if len(suffix) == 0:
            setattr(self, prefix, value)
        else:
            getattr(self, prefix)[suffix] = value


class SymbolExtractor(ast.NodeVisitor):
    def __init__(self):
        self.symbols = set()
        self.attribute_accesses = set()

    def visit_Name(self, node):
        """Extract variable names"""
        if isinstance(node.ctx, ast.Load):  # Only interested in reads, not assignments
            self.symbols.add(node.id)
        self.generic_visit(node)

    def visit_Attribute(self, node):
        """Extract attribute accesses like model.backbone.name"""
        if isinstance(node.ctx, ast.Load):  # Only interested in reads
            # Build the full attribute chain
            parts = []
            current = node
            while isinstance(current, ast.Attribute):
                parts.append(current.attr)
                current = current.value
            if isinstance(current, ast.Name):
                parts.append(current.id)
                full_path = ".".join(reversed(parts))
                self.attribute_accesses.add(full_path)
        self.generic_visit(node)


def extract_symbols_from_code(code_string: str):
    """
    Extract all symbols and attribute accesses from Python code.
    Returns: (simple_symbols, attribute_chains)
    """
    try:
        tree = ast.parse(code_string)
        extractor = SymbolExtractor()
        extractor.visit(tree)
        return extractor.symbols, extractor.attribute_accesses
    except SyntaxError as e:
        print(f"Syntax error in code: {e}")
        return set(), set()



from pathlib import Path
from typing import Union

import nip.elements as elements
from nip.stream import Stream


class Parser:
    def __init__(
        self,
        implicit_fstrings: bool = True,
        strict: bool = False,
        sequential_links: bool = False,
    ):
        self.links = []
        self.iterators = []
        self.link_replacements = {}  # used with !!insert directive
        self.implicit_fstrings = implicit_fstrings
        self.strict = strict
        self.sequential_links = sequential_links
        self.last_indent = -1
        self.stack = []

    def parse(self, path: Union[str, Path]):
        path = Path(path)
        with path.open() as f_stream:
            string_representation = f_stream.read()

        tree = self.parse_string(string_representation)
        tree._path = path
        return tree

    def __call__(self, path: Union[str, Path]):
        return self.parse(path)

    def parse_string(self, string):
        stream = Stream(string)  # mb: add stream to parser and log errors more convenient
        tree = elements.Document.read(stream, self)
        if stream:
            raise ParserError(stream, "Wrong statement.")
        tree._update_parents()
        return tree

    def has_iterators(self) -> bool:
        return len(self.iterators) > 0


class ParserError(Exception):
    def __init__(self, stream: Stream, msg: str):
        self.line = stream.line
        self.pos = stream.pos
        self.msg = msg

    def __str__(self):
        return f"{self.line + 1}:{self.pos + 1}: {self.msg}"



from .constructor import Constructor
from .constructor import nip, wrap_module
from .convertor import pin
from .elements import Node
from .main import (
    parse,
    parse_string,
    construct,
    load,
    load_string,
    dump,
    dump_string,
    convert,
    run,
    update,
    update_flatten,
)
from .non_seq_constructor import NonSequentialConstructor



"""Contains all the elements of nip config files"""

from __future__ import annotations

import logging
import re
from abc import abstractmethod, ABC
from pathlib import Path
from typing import Any, Union, Tuple, Dict

import nip.constructor  # This import pattern because of cycle imports
import nip.directives
import nip.dumper
import nip.non_seq_constructor as nsc
import nip.parser
import nip.stream
import nip.tokens as tokens
import nip.utils
import nip.dict

_LOGGER = logging.getLogger(__name__)


class Node(ABC, object):
    def __init__(self, name: str = "", value: Any = None, line: int = None, pos: int = 0):
        self._name = name
        self._value = value
        self._parent = None
        self._line = line
        self._pos = pos

    @classmethod
    @abstractmethod
    def read(cls, stream: nip.stream.Stream, parser: nip.parser.Parser) -> Union[Node, None]:
        pass

    def __str__(self):
        return f"{self.__class__.__name__}('{self._name}', {self._value})"

    def __getitem__(self, item):
        if not isinstance(item, (str, int)):
            raise TypeError(f"Unexpected item type: {type(item)}. str or int are expected.")
        if isinstance(item, str) and len(item) == 0:
            return self
        if self._value is None:
            raise KeyError(f"'{item}' is not a part of the Node.")
        return self._value[item]

    def __getattr__(self, item):  # unable to access names like `construct` and 'dump` via this method
        return self.__getitem__(item)

    def __setitem__(self, key, value):
        self._value[key] = value
        self._value._parent = self

    def __contains__(self, item):
        if not isinstance(self._value, Node):
            return False
        return item in self._value

    def __setattr__(self, key, value):
        if key.startswith("_"):  # mb: ensure not user's node name?
            self.__dict__[key] = value
        else:
            self.__setitem__(key, value)

    def to_python(self):
        return self._value.to_python()

    def to_dictobject(self):
        data = self.construct()
        if isinstance(data, (tuple, list, dict)):
            return nip.dict.DictObject(data)
        return data

    @nip.constructor.construct_method
    def _construct(self, constructor: nip.constructor.Constructor):
        return self._value._construct(constructor)

    def construct(self, strict_typing: bool = False, as_dictobj: bool = False):
        return nip.construct(self, strict_typing=strict_typing, nonsequential=True, as_dictobj=as_dictobj)

    def _dump(self, dumper: nip.dumper.Dumper):
        return self._value._dump(dumper)

    def dump(self, path: Union[str, Path]):
        nip.dump(path, self)

    def dump_string(self):
        return nip.dump_string(self)

    def __eq__(self, other):
        return self._name == other._name and self._value == other._value

    def flatten(self, delimiter=".") -> Dict:
        return nip.utils.flatten(self.to_python(), delimiter)

    def _update_parents(self):
        if isinstance(self._value, Node):
            self._value._parent = self
            self._value._update_parents()

    def _get_root(self):
        if self._parent is None:
            return self
        return self._parent._get_root()

    def update(self):
        self._get_root().update()


Element = Node  # backward compatibility until 1.* version


class Document(Node):  # ToDo: add multi document support
    def __init__(self, name: str = "", value: Union[Node, Any] = None, line: int = None, pos: int = 0):
        super().__init__(name, value)
        self._path = None
        self._line = line
        self._pos = pos

    @classmethod
    def read(cls, stream: nip.stream.Stream, parser: nip.parser.Parser) -> Document:
        line, pos = stream.line, stream.pos
        doc_name = cls._read_name(stream)
        content = read_node(stream, parser)
        return Document(doc_name, content, line, pos)

    @classmethod
    def _read_name(cls, stream: nip.stream.Stream):
        read_tokens = stream.peek(tokens.Operator("---"), tokens.Name) or stream.peek(tokens.Operator("---"))
        if read_tokens is not None:
            line, pos = stream.step()
            if len(read_tokens) == 2:
                return read_tokens[1]._value
        return ""

    def _dump(self, dumper: nip.dumper.Dumper):
        string = "---"
        if self._name:
            string += " " + self._name + " "
        return string + self._value._dump(dumper)

    def update(self):
        self.dump(self._path)


class Value(Node):
    @classmethod
    def read(cls, stream: nip.stream.Stream, parser: nip.parser.Parser) -> Union[None, Value]:
        tokens_list = [
            tokens.Number,
            tokens.NoneToken,
            tokens.Bool,
            tokens.String,
            tokens.List,
            tokens.TupleToken,
            tokens.Dict,
        ]
        for token in tokens_list:
            read_tokens = stream.peek(token)
            if read_tokens is not None:
                line, pos = stream.step()
                return Value(read_tokens[0]._name, read_tokens[0]._value, line=line, pos=pos)

        return None

    def to_python(self):
        return self._value

    @nip.constructor.construct_method
    def _construct(self, constructor: nip.constructor.Constructor = None):
        constructor[self] = self._value
        return self._value

    def _dump(self, dumper: nip.dumper.Dumper):
        if isinstance(self._value, str):
            return f'"{self._value}"'
        return str(self._value)

    def __len__(self):
        return len(self._value)


class LinkCreation(Node):
    @classmethod
    def read(cls, stream: nip.stream.Stream, parser: nip.parser.Parser) -> Union[Node, None]:
        read_tokens = stream.peek(tokens.Operator("&"), tokens.Name)
        if read_tokens is None:
            # if stream.peek(tokens.Operator('&')):  # mb: do it more certainly: peak operator
            #     raise nip.parser.ParserError(      # mb: firstly and then choose class to read)
            #         stream, "Found variable creation operator '&' but name is not specified")
            return None

        name = read_tokens[1]._value
        line, pos = line, pos = stream.step()

        value = read_node(stream, parser)
        if name in parser.links:
            raise nip.parser.ParserError(stream, f"Redefining of link '{name}'")
        parser.links.append(name)

        return LinkCreation(name, value, line, pos)

    @nip.constructor.construct_method
    def _construct(self, constructor: nip.constructor.Constructor):
        constructor[self._name] = self
        return self._value._construct(constructor)

    def _dump(self, dumper: nip.dumper.Dumper):
        return f"&{self._name} {self._value._dump(dumper)}"


class Link(Node):
    @classmethod
    def read(cls, stream: nip.stream.Stream, parser: nip.parser.Parser) -> Union[Node, None]:
        read_tokens = stream.peek(tokens.Operator("*"), tokens.Name)
        if read_tokens is None:
            return None

        name = read_tokens[1]._value  # mb: use LinkCreation node as value. fixes `in` operator.
        line, pos = stream.step()

        if name in parser.link_replacements:
            return parser.link_replacements[name]

        if parser.sequential_links and name not in parser.links:
            nip.parser.ParserError(stream, "Link usage before assignment")

        return Link(name, line=line, pos=pos)

    def to_python(self):
        return "nil"  # something that means that object is not constructed yet.

    @nip.constructor.construct_method
    def _construct(self, constructor: nip.non_seq_constructor.NonSequentialConstructor):
        # if self._construction_in_progress:
        #     raise nip.non_seq_constructor.NonSequentialConstructorError(f"Recursive construction of {self._name}")
        # self._construction_in_progress = True
        root = self._get_root()
        if self._name in constructor.links:  # was constructed or can be constructed
            value = constructor[self._name]
        elif self._name in root:
            value = root[self._name]._construct(constructor)
        else:
            raise NameError(f"Unable to resolve link '{self._name}'.")
        # self._construction_in_progress = False
        return value

    def _dump(self, dumper: nip.dumper.Dumper):
        return f"*{self._name}"

    def __getitem__(self, item):
        raise NotImplementedError("'__getitem__' is not implemented for Link node.")

    def __contains__(self, item):
        raise NotImplementedError("'in' operator if not implemented for Link node.")


class Tag(Node):
    @classmethod
    def read(cls, stream: nip.stream.Stream, parser: nip.parser.Parser) -> Union[Tag, None]:
        read_tokens = stream.peek(tokens.Operator("!"), tokens.Name)
        if read_tokens is None:
            return None
        name = read_tokens[1]._value
        line, pos = stream.step()

        value = read_node(stream, parser)

        return Tag(name, value, line=line, pos=pos)

    @nip.constructor.construct_method
    def _construct(self, constructor: nip.constructor.Constructor):
        if isinstance(self._value, Args):
            args, kwargs = self._value._construct(constructor, always_pair=True)
        else:
            value = self._value._construct(constructor)
            if isinstance(value, Nothing):  # mb: Add IS_NOTHING method
                args, kwargs = [], {}
            else:
                args, kwargs = [value], {}

        return nip.constructor.construct_with_args(self._name, args, kwargs, constructor, self)

    def _dump(self, dumper: nip.dumper.Dumper):
        return f"!{self._name} " + self._value._dump(dumper)


class Class(Node):
    @classmethod
    def read(cls, stream: nip.stream.Stream, parser: nip.parser.Parser) -> Union[Class, None]:
        read_tokens = stream.peek(tokens.Operator("!&"), tokens.Name)
        if read_tokens is None:
            return None
        name = read_tokens[1]._value
        line, pos = stream.step()

        value = read_node(stream, parser)
        if not isinstance(value, Nothing):
            raise nip.parser.ParserError(stream, "Class should be created with nothing to the right.")

        return Class(name, value, line=line, pos=pos)

    @nip.constructor.construct_method
    def _construct(self, constructor: nip.constructor.Constructor):
        value = self._value._construct(constructor)
        assert isinstance(value, Nothing), "Unexpected right value while constructing Class"

        return constructor.builders[self._name]

    def _dump(self, dumper: nip.dumper.Dumper):
        return f"!&{self._name} " + self._value._dump(dumper)


class Args(Node):
    def __init__(self, args, kwargs, name: str = "", line=None, pos=None):
        self._name = name
        self._args = args
        self._kwargs = kwargs
        self._value = None  # to prevent step into
        self._parent = None
        self._line = line
        self._pos = pos

    @classmethod
    def read(cls, stream: nip.stream.Stream, parser: nip.parser.Parser) -> Union[Args, None]:
        start_indent = stream.pos
        if start_indent <= parser.last_indent:
            return None

        args = []
        kwargs = {}  # mb: just dict with integer and string keys !
        read_kwarg = False
        pos, read_pos = [None, None], None
        while stream and stream.pos == start_indent:
            pos = pos or read_pos
            parser.last_indent = start_indent

            item, read_pos = cls._read_list_item(stream, parser)
            if item is not None:
                if parser.strict and read_kwarg:
                    raise nip.parser.ParserError(
                        stream,
                        "Positional argument after keyword argument is forbidden in `strict` mode.",
                    )
                args.append(item)
                continue

            key, value, read_pos = cls._read_dict_pair(stream, parser, kwargs.keys())
            if key is not None:
                read_kwarg = True
                kwargs[key] = value
                continue

            break

        if stream.pos > start_indent:
            raise nip.parser.ParserError(stream, "Unexpected indent")

        if not args and not kwargs:
            return None
        pos = pos or (None, None)
        return Args(args, kwargs, "args", line=pos[0], pos=pos[1])

    @classmethod
    def _read_list_item(
        cls, stream: nip.stream.Stream, parser: nip.parser.Parser
    ) -> Union[Tuple[Node, Tuple[int, int]], Tuple[None, None]]:
        read_tokens = stream.peek(tokens.Operator("- "))
        if read_tokens is None:
            return None, None
        line, pos = stream.step()

        value = read_node(stream, parser)

        return value, (line, pos)

    @classmethod
    def _read_dict_pair(
        cls, stream: nip.stream.Stream, parser: nip.parser.Parser, kwargs_keys
    ) -> Union[Tuple[str, Node, Tuple[int, int]], Tuple[None, None, None]]:
        # mb: read String instead of Name for keys with spaces,
        # mb: but this leads to the case that
        read_tokens = stream.peek(tokens.Name, tokens.Operator(": "))
        if read_tokens is None:
            return None, None, None

        key = read_tokens[0]._value
        if parser.strict and key in kwargs_keys:
            raise nip.parser.ParserError(
                stream,
                f"Dict key overwriting is forbidden in `strict` " f"mode. Overwritten key: '{key}'.",
            )
        line, pos = stream.step()

        value = read_node(stream, parser)

        return key, value, (line, pos)

    def __str__(self):
        args_repr = "[" + ", ".join([str(item) for item in self._args]) + "]"
        kwargs_repr = "{" + ", ".join([f"{key}: {str(value)}" for key, value in self._kwargs.items()]) + "}"

        return f"{self.__class__.__name__}('{self._name}', {args_repr}, {kwargs_repr})"

    def __bool__(self):  # mb: should always be True.
        return bool(self._args) or bool(self._kwargs)

    def _is_list(self):
        return len(self._args) > 0 and len(self._kwargs) == 0

    def _is_dict(self):
        return len(self._args) == 0 and len(self._kwargs) > 0

    def _is_args(self):
        return len(self._args) > 0 and len(self._kwargs) > 0

    def _get_sub_item(self, item):
        # some.deep.0.parameter -> [some.deep][0][parameter] / [some][deep][0][parameter]
        if not isinstance(item, (str, int)):
            raise TypeError(f"Unexpected item type: {type(item)}. str or int are expected.")

        if isinstance(item, int) or item.isnumeric():
            item = int(item)
            if 0 <= item < len(self._args):
                return None, self._args[item]
            return None, None
        key = item.split(".")[0]
        if key.isnumeric():
            # left_key = None if len(key) == len(item) else item[len(key) + 1 :]
            return item[len(key) + 1 :], self._args[int(key)]
        for key in self._kwargs:
            if item.startswith(key):
                if len(item) == len(key):
                    return None, self._kwargs[key]
                if item[len(key)] != ".":
                    continue
                return item[len(key) + 1 :], self._kwargs[key]
        return None, None

    def _set_sub_item(self, key, value):
        if isinstance(key, int) or key.isnumeric():
            key = int(key)
            if 0 <= key < len(self._args):
                self._args[key] = value
            elif key == len(self._args):
                self._args.append(value)
            else:
                raise KeyError("You may only update existing arg of the Node or add one using `len(args)` as index")
        else:
            self._kwargs[key] = value

    def __getitem__(self, item) -> Node:
        left_key, node = self._get_sub_item(item)
        if node is None:
            raise KeyError(f"'{item}' is not a part of the Node.")
        if left_key:  # step deeper
            return node[left_key]
        return node

    def __contains__(self, item):
        left_key, node = self._get_sub_item(item)
        if node is None:
            return False
        if left_key:  # step deeper
            return left_key in node
        return True

    def __setitem__(self, key, value):
        if not isinstance(value, Node):
            value = nip.convert(value)
        if isinstance(value, Document):  # this convenient for user. but do not insert Document node inside the tree.
            value = value._value

        left_key, node = self._get_sub_item(key)
        if node is None:  # new sub
            self._set_sub_item(key, value)
            return
        if left_key:  # step deeper
            node[left_key] = value
        else:  # update sub
            self._set_sub_item(key, value)

    def append(self, value):
        self._args.append(nip.convert(value))

    def __len__(self):
        return len(self._args) + len(self._kwargs)

    def __iter__(self):
        for i, item in enumerate(self._args):
            yield i, item
        for key, item in self._kwargs.items():
            yield key, item

    def to_python(self):
        args = list(item.to_python() for item in self._args)
        kwargs = {key: value.to_python() for key, value in self._kwargs.items()}
        assert args or kwargs, "Error converting Args node to python."  # This should never happen
        if args and kwargs:
            result = {}
            result.update(nip.utils.iterate_items(args))
            result.update(nip.utils.iterate_items(kwargs))
            return result
        return args or kwargs

    @nip.constructor.construct_method
    def _construct(self, constructor: nip.constructor.Constructor, always_pair=False):
        args = list(item._construct(constructor) for item in self._args)
        kwargs = {
            key: value._construct(constructor)
            for key, value in self._kwargs.items()
            if key not in ["_target_", "_args_"]
        }
        if "_target_" in self._kwargs:
            name = self._kwargs["_target_"]._construct(constructor)
            if "_args_" in self._kwargs:
                if len(args) > 0:
                    nip.constructor.ConstructorError(
                        self,
                        args,
                        kwargs,
                        "'_args_' key and usual args cant be presented in the Node at the same time.",
                        name=name,
                    )
                args = self._kwargs["_args_"]._construct(constructor)
            return nip.constructor.construct_with_args(name, args, kwargs, constructor, self)
        assert args or kwargs, "Error constructing Args node."  # This should never happen
        if args and kwargs or always_pair:
            return args, kwargs
        # return args or kwargs
        return args or (nip.dict.DictObject(kwargs) if constructor.as_dictobj else kwargs)

    def _dump(self, dumper: nip.dumper.Dumper):
        dumped_args = "\n".join(
            [" " * dumper.indent + f"- {item._dump(dumper + dumper.default_shift)}" for item in self._args]
        )
        string = ("\n" if dumped_args else "") + dumped_args

        dumped_kwargs = "\n".join(
            [
                " " * dumper.indent + f"{key}: {value._dump(dumper + dumper.default_shift)}"
                for key, value in self._kwargs.items()
            ]
        )
        string += ("\n" if dumped_kwargs else "") + dumped_kwargs

        return string

    def _update_parents(self):
        self.__dict__.update(self._kwargs)
        for key, item in self:
            item._parent = self
            item._update_parents()


class Iter(Node):  # mark all parents as Iterable and allow construct specific instance
    def __init__(self, name: str = "", value: Any = None, line: int = None, pos: int = None):
        super(Iter, self).__init__(name, value)
        self._return_index = -1
        # mb: name all the iterators and get the value from constructor rather then use this index
        self._line = line
        self._pos = pos

    @classmethod
    def read(cls, stream: nip.stream.Stream, parser: nip.parser.Parser) -> Union[Iter, None]:
        read_tokens = stream.peek(tokens.Operator("@"), tokens.Name) or stream.peek(tokens.Operator("@"))
        if read_tokens is None:
            return None
        line, pos = stream.step()
        value = read_node(stream, parser)
        if isinstance(value, Value) and isinstance(value._value, list):  # mb: replace inline list with Node
            value = value._value
        elif isinstance(value, Args) and value._is_list():
            value = value
        else:
            raise nip.parser.ParserError(stream, "List is expected as a value for Iterable node")
        if len(read_tokens) == 1:
            iterator = Iter("", value, line=line, pos=pos)
        else:
            iterator = Iter(read_tokens[1]._value, value, line=line, pos=pos)

        parser.iterators.append(iterator)
        return iterator

    def to_python(self):
        if self._return_index == -1:
            raise iter(self._value)
        if isinstance(self._value[self._return_index], Node):
            return self._value[self._return_index].to_python()
        return self._value[self._return_index]

    @nip.constructor.construct_method
    def _construct(self, constructor: nip.constructor.Constructor):
        if self._return_index == -1:
            raise Exception("Iterator index was not specified by IterParser")
        if isinstance(self._value, list):
            return self._value[self._return_index]
        elif isinstance(self._value, Args):
            return self._value[self._return_index]._construct(constructor)
        else:
            raise nip.constructor.ConstructorError(self, (), {}, "Unexpected iter value type")

    def _dump(self, dumper: nip.dumper.Dumper):
        if self._return_index == -1:
            raise nip.dumper.DumpError("Dumping an iterator but index was not specified by IterParser")
        if isinstance(self._value, list):
            return str(self._value[self._return_index])
        elif isinstance(self._value, Args):
            return self._value[self._return_index]._dump(dumper)
        else:
            raise nip.dumper.DumpError("Unable to dump Iterable node: unexpected value type")


class InlinePython(Node):
    @classmethod
    def read(cls, stream: nip.stream.Stream, parser: nip.parser.Parser) -> Union[InlinePython, None]:
        read_tokens = stream.peek(tokens.InlinePython)
        if read_tokens is None:
            return None
        line, pos = stream.step()
        exec_string = read_tokens[0]._value
        return InlinePython(value=exec_string, line=line, pos=pos)

    @nip.constructor.construct_method
    def _construct(self, constructor: nip.constructor.Constructor):
        symbols, attributes_access = nip.utils.extract_symbols_from_code(self._value)
        namespace = nip.utils.Namespace()
        root = self._get_root()
        for item in attributes_access:
            if item in root:
                namespace[item] = constructor[root[item]]  # != root[item]._construct(constructor) for Constructor
        for symbol in symbols:
            if symbol in constructor:
                namespace[symbol] = constructor[symbol]
        # mb: check we didn't find attributes. default exception is fine?
        # but we need to do this for symbols that are not links
        locals().update(namespace.__dict__)
        return eval(self._value)

    def _dump(self, dumper: nip.dumper.Dumper):
        return f"`{self._value}`"

    def to_python(self):
        return f"`{self._value}`"


class Nothing(Node):
    @classmethod
    def read(cls, stream: nip.stream.Stream, parser: nip.parser.Parser) -> Union[Nothing, None]:
        line, pos = stream.line, stream.pos
        if not stream:
            return Nothing(line=line, pos=pos)

        indent = stream.pos
        if stream.pos == 0 or (stream.lines[stream.line][: stream.pos].isspace() and indent <= parser.last_indent):
            return Nothing(line=line, pos=pos)

    @nip.constructor.construct_method
    def _construct(self, constructor: nip.constructor.Constructor):
        return self

    def _dump(self, dumper: nip.dumper.Dumper):
        return ""

    def to_python(self):
        return None


class FString(Node):  # Includes f-string and r-string
    @classmethod
    def read(cls, stream: nip.stream.Stream, parser: nip.parser.Parser) -> Union[FString, None]:
        read_tokens = stream.peek(tokens.PythonString)
        if read_tokens is None:
            return None
        line, pos = stream.step()
        string, t = read_tokens[0]._value
        if t == "r":
            print(
                "Warning: all strings in NIP are already python r-string. " "You don't have to explicitly specify it."
            )
        return FString(value=string, line=line, pos=pos)

    @nip.constructor.construct_method
    def _construct(self, constructor: nip.constructor.Constructor):
        symbols, attributes_access = nip.utils.extract_symbols_from_code(f"f{self._value}")
        namespace = nip.utils.Namespace()
        root = self._get_root()
        for item in attributes_access:
            if item in root:
                namespace[item] = constructor[root[item]]  # != root[item]._construct(constructor) for Constructor
        for symbol in symbols:
            if symbol in constructor:
                namespace[symbol] = constructor[symbol]
        # mb: check we didn't find attributes. default exception is fine?
        # but we need to do this for symbols that are not links
        locals().update(namespace.__dict__)
        return eval(f"f{self._value}")

    def _dump(self, dumper: nip.dumper.Dumper):
        return f"f{self._value}"

    def to_python(self):
        return f"f{self._value}"


class Directive(Node):
    @classmethod
    def read(cls, stream: nip.stream.Stream, parser: nip.parser.Parser) -> Union[FString, None]:
        read_tokens = stream.peek(tokens.Operator("!!"), tokens.Name)
        if read_tokens is None:
            return None
        name = read_tokens[1]._value
        stream.step()

        value = read_node(stream, parser)

        return nip.directives.call_directive(name, value, stream)


def read_node(stream: nip.stream.Stream, parser: nip.parser.Parser) -> Node:
    value = (
        Directive.read(stream, parser)
        or LinkCreation.read(stream, parser)
        or Link.read(stream, parser)
        or Class.read(stream, parser)
        or Tag.read(stream, parser)
        or Iter.read(stream, parser)
        or Args.read(stream, parser)
        or FString.read(stream, parser)
        or Nothing.read(stream, parser)
        or InlinePython.read(stream, parser)
        or Value.read(stream, parser)
    )

    if value is None:
        raise nip.parser.ParserError(stream, "Wrong right value")

    return value



import symtable

import nip.elements
from .constructor import Constructor
from typing import Dict, Any


class NonSequentialConstructor(Constructor):
    def __init__(
        self,
        base_config: "nip.elements.Node",
        ignore_rewriting=False,
        load_builders=True,
        strict_typing=False,
        as_dictobj: bool = False,
    ):
        super().__init__(ignore_rewriting, load_builders, strict_typing, as_dictobj=as_dictobj)
        self.links = {}  # name -> node
        self._find_links(base_config)
        self.constructed_nodes = {}  # node -> obj  # store here because of gc
        self.in_progress = set()

    def _find_links(self, node: "nip.elements.Node"):
        if isinstance(node, nip.elements.LinkCreation):
            assert node._name not in self.links, "Redefined link."
            self.links[node._name] = node
        if isinstance(node, nip.elements.Args):
            for key, sub_node in node:
                self._find_links(sub_node)
        if isinstance(node._value, nip.elements.Node):
            self._find_links(node._value)

    def __contains__(self, item):
        assert isinstance(item, (str, nip.elements.Node))
        if isinstance(item, str):
            return item in self.links
        return id(item) in self.constructed_nodes

    def __setitem__(self, key, value):
        if isinstance(key, str):  # we already found all the links for nonseq
            return
        assert isinstance(key, nip.elements.Node)
        self.constructed_nodes[id(key)] = value

    def __getitem__(self, item):
        if isinstance(item, str):
            if item not in self.links:
                raise NonSequentialConstructorError(f"Unresolved reference '{item}'")  # mb: link, ref or var
            item = self.links[item]
        assert isinstance(item, nip.elements.Node)
        if id(item) in self.in_progress:
            raise NonSequentialConstructorError(f"Recursive construction of '{item}'.")
        if id(item) not in self.constructed_nodes:
            self.in_progress.add(id(item))
            self.constructed_nodes[id(item)] = item._construct(self)
            self.in_progress.remove(id(item))
        return self.constructed_nodes[id(item)]


class NonSequentialConstructorError(Exception):
    def __init__(self, massage):
        self.massage = massage

    def __str__(self):
        return self.massage


def preload_vars(code, constructor: Constructor):
    if not isinstance(constructor, NonSequentialConstructor):
        return
    table = symtable.symtable(code, "string", "exec")
    for name in constructor.links:
        try:
            if table.lookup(name).is_global():
                constructor.links[name]
        except KeyError:
            pass



from typing import Union, Type

import nip.tokens as tokens


class Stream:
    def __init__(self, sstream: str):
        self.lines = sstream.split("\n")
        self.lines = [line + " " for line in self.lines]
        self.line = 0
        self.pos = 0
        self.last_peak_pos = -1
        self._pass_forward()

    def peek(self, *args: Union[tokens.Token, Type[tokens.Token]]):
        """Reads several tokens from stream"""
        if not self:
            return None
        line = self.lines[self.line]
        pos = self.pos
        self.last_peak_pos = -1  # prevent step() after failed peek()
        read_tokens = []
        for arg in args:
            if isinstance(arg, tokens.Token):
                token_type = arg.__class__
            else:
                token_type = arg

            # skip empty line ending
            while pos < len(line) and line[pos].isspace():
                pos += 1
            if pos >= len(line):
                return None

            try:
                length, token = token_type.read(line[pos:])
                # mb: pass full stream to token. (This will allow multiline string parsing)
            except tokens.TokenError as e:
                raise StreamError(self.line, pos, e)

            if token is None:
                return None
            if isinstance(arg, tokens.Token) and token != arg:
                return None

            token.set_position(self.line, pos)
            read_tokens.append(token)
            pos += length

        self.last_peak_pos = pos
        return read_tokens

    def step(self):
        assert self.last_peak_pos > 0, "step() called before peaking any Token"
        line, pos = self.line, self.pos
        self.pos = self.last_peak_pos
        self._pass_forward()
        return line, pos  # the point we started reading, since this log is more convenient for user

    def _pass_forward(self):
        while self and (
            self.pos >= len(self.lines[self.line])
            or self.lines[self.line][self.pos :].isspace()
            or self.lines[self.line][self.pos :].strip()[0] == "#"
        ):
            self.line += 1
            self.pos = 0

        if not self:
            return

        while self.lines[self.line][self.pos].isspace():
            self.pos += 1

    def __bool__(self):
        return self.line < len(self.lines)


class StreamError(Exception):
    def __init__(self, line: int, position: int, msg: Exception):
        self.line = line
        self.pos = position
        self.msg = msg

    def __str__(self):
        return f"{self.line + 1}:{self.pos + 1}: {self.msg}"



from typing import List, Any, Union

from .utils import iterate_items


class DictObject(dict):
    def __getattr__(self, item):
        return self[item]

    def add_item(self, key, value):
        prefix = key.split(".")[0]
        suffix = ".".join(key.split(".")[1:])
        if len(suffix) == 0:
            setattr(self, prefix, value)
        else:
            getattr(self, prefix)[suffix] = value



# Constructor of tagged objects
import functools
import importlib
import importlib.util
import logging
import pydoc
from types import FunctionType, ModuleType, BuiltinFunctionType
from typing import Callable, Optional, Union

from .utils import get_sub_dict, check_typing

global_builders = {}  # builders shared between Constructors
global_calls = {}  # history of object creations

_LOGGER = logging.getLogger(__name__)


class Constructor:
    def __init__(self, ignore_rewriting=False, load_builders=True, strict_typing=False, as_dictobj: bool = False):
        self.builders = {}
        self.ignore_rewriting = ignore_rewriting
        if load_builders:
            self.load_builders()
        self.links = {}
        self.strict_typing = strict_typing
        self.constructed_nodes = {}
        self.as_dictobj = as_dictobj

    def construct(self, element):
        return element._construct(self)

    def register(self, func: Callable, tag: Optional[str] = None):
        """Registers builder function for tag.

        Parameters
        ----------
        func:
            Function or class to build the python object.
            In case of class its __init__ method will be called to construct object.
        tag: str, optional
            Tag in yaml/nip file. func.__name__ will be used if not specified.
        """
        if tag is None:
            tag = func.__name__
        assert self.ignore_rewriting or tag not in self.builders, f"Builder for tag '{tag}' already registered"
        self.builders[tag] = func

    def load_builders(self):
        self.builders.update(global_builders)
        self.builders.update(get_sub_dict(NIPBuilder))

    def __contains__(self, item):
        if isinstance(item, str):
            return item in self.links
        return id(item) in self.constructed_nodes

    def __getitem__(self, item):
        if isinstance(item, str):
            item = self.links[item]
        return self.constructed_nodes[id(item)]

    def __setitem__(self, key, value):
        if isinstance(key, str):  # create link
            self.links[key] = value
        self.constructed_nodes[id(key)] = value  # constructed node


class ConstructorError(Exception):
    def __init__(self, node, args, kwargs, e, name=None):
        self.cls = type(node).__name__
        self.line = node._line
        self.pos = node._pos
        self.name = name or node._name
        self.args = args
        self.kwargs = kwargs
        self.e = e

    def __str__(self):
        return (
            f"Unable to construct {self.cls} '{self.name}' at {self.line}:{self.pos} with args:{self.args} and "
            f"kwargs:{self.kwargs}.\nFollowing exception occurred:\n"
            f"{self.e.__class__.__name__}: {self.e}"
        )


def construct_with_args(name, args, kwargs, constructor: Constructor, node):
    builder = None
    if name in constructor.builders:
        builder = constructor.builders[name]
    builder = builder or pydoc.locate(name)
    if builder is None:
        raise ConstructorError(
            node, args, kwargs, f"Builder for '{name}' is not registered and unable to locate.", name=name
        )

    messages = check_typing(builder, args, kwargs)
    if len(messages) > 0:
        if constructor.strict_typing:
            raise ConstructorError(node, args, kwargs, "\n".join(messages), name=name)
        else:
            _LOGGER.warning(f"Typing mismatch while constructing {name}:\n" + "\n".join(messages))

    try:  # Try to construct
        return builder(*args, **kwargs)
    except Exception as e:
        raise ConstructorError(node, args, kwargs, e, name=name)


def construct_method(method: Callable) -> Callable:
    @functools.wraps(method)
    def wrapper(self, constructor, *args, **kwargs):
        if self in constructor:
            return constructor[self]
        result = method(self, constructor, *args, **kwargs)
        constructor[self] = result
        return result

    return wrapper


# mb: add meta for auto detecting this class as NIP-builder
# ToDo: Add init wrapper for auto detection init args for convenient object dumping
class NIPBuilder:
    pass


def nip_decorator(name=None, convertable=False):
    assert name is None or len(name) > 0, "name should be nonempty"

    def _(item):
        if convertable:
            assert isinstance(item, type), "Call wrapping supported only for class type"
            make_convertable(item)
        local_name = name or item.__name__
        if isinstance(local_name, (list, tuple)):
            for n in local_name:
                global_builders[n] = item
        else:
            global_builders[local_name] = item
        return item

    return _


# instead of multipledispatch
def nip(item=None, *, wrap_builtins=False, convertable=False):
    if isinstance(item, str):  # single name is passed
        return nip_decorator(item, convertable)
    if isinstance(item, (list, tuple)):
        for name in item:
            if not isinstance(name, str):
                raise ValueError("Every specified Tag should be a string.")
        return nip_decorator(item, convertable)
    if isinstance(item, (type, FunctionType, BuiltinFunctionType)):
        return nip_decorator(convertable=convertable)(item)
    if isinstance(item, ModuleType):
        return wrap_module(item, wrap_builtins=wrap_builtins, convertable=convertable)
    if item is not None:
        raise ValueError("Unexpected type passed to @nip decorator.")
    return nip_decorator(convertable=convertable)


def wrap_module(module: Union[str, ModuleType], wrap_builtins=False, convertable=False):
    """Wraps everything declared in module with @nip

    Parameters
    ----------
    module: str or ModuleType
        Module name (e.g. "numpy.random") or module itself
    wrap_builtins
        Whether to wrap builtin functions or not.
        (Useful when wrapping whole module like `numpy`)
    """
    if isinstance(module, str):
        module = importlib.import_module(module)

    for value in module.__dict__.values():
        if isinstance(value, (type, FunctionType)) or wrap_builtins and isinstance(value, BuiltinFunctionType):
            nip(value, convertable=convertable and isinstance(value, type))

    return module


class ArgsKwargs:
    def __init__(self, args, kwargs):
        self.args = args
        self.kwargs = kwargs


def _wrap_init_call(self, *args, **kwargs):
    self.__init_args = ArgsKwargs(args, kwargs)
    self.__origin_init__(*args, **kwargs)


def _converter(self):
    return self.__init_args


def make_convertable(cls):
    if hasattr(cls, "__nip__"):
        return
    cls.__origin_init__ = cls.__init__
    cls.__init__ = _wrap_init_call
    cls.__nip__ = _converter
